from types import FunctionType, LambdaType, MethodType, BuiltinFunctionType, BuiltinMethodType

class luaArgList():
	def __init__(self, values):
		self.__values = values.copy()
	
	def getValue(self):
		return self.__values

def argumentValidation(args):
	if isinstance(args, luaArgList):
		args = args.getValue()
	elif len(args) > 0:
		if isinstance(args[0], luaArgList):
			args = args[0].getValue()
	
	argList = []
	for arg in args:
		if isinstance(arg, list):
			# convert list to dict
			arg = dict(zip(range(1, len(arg)+1), arg))
		argList.append(arg)
	
	return argList

def toBool(value):
	if isinstance(value, bool):
		return value
	elif value == None:
		return False
	else:
		return True

def toNum(value):
	if isinstance(value, str):
		if "." in value:
			return float(value)
		else:
			return int(value)
	else:
		return value

def toString(value):
	type = luaType(value)
	if type == "string":
		return value
	elif type == "number":
		if int(value) == value:
			return str(int(value))
		return str(value)
	elif type == "bool":
		if value:
			return "true"
		else:
			return "false"
	else:
		return type

def luaType(value):
	if isinstance(value, (int, float)):
		return "number"
	elif isinstance(value, str):
		return "string"
	elif isinstance(value, bool):
		return "boolean"
	elif value == None:
		return "nil"
	elif isinstance(value, dict):
		return "table"
	elif isinstance(value, (FunctionType, LambdaType, MethodType, BuiltinFunctionType, BuiltinMethodType)):
		return "function"
	else:
		return str(type(value))

def _ipairs(table):
	if not isinstance(table, dict):
		raise ValueError("attempt to index a %s value" % luaType(table))
	else:
		i = 1
		pairs = []
		while i in table:
			pairs.append((i, table[i]))
			i += 1
		return iter(pairs)

def _pairs(table):
	if not isinstance(table, dict):
		raise ValueError("bad argument #1 to 'for iterator' (table expected, got %s)" % luaType(table))
	else:
		return table.items()

def _error(message):
	if not isinstance(message, str):
		raise ValueError("bad argument #1 to 'error' (string expected, got %s)" % luaType(message))
	else:
		raise RuntimeError(message)
		
globalVars = {
	"print": lambda *args: print(argumentValidation(args)),
	"ipairs": lambda *args: _ipairs(argumentValidation(args)[0]),
	"pairs": lambda *args: _pairs(argumentValidation(args)[0]),
	"error": lambda *args: _error(argumentValidation(args)[0]),
	"type": lambda *args: luaType(argumentValidation(args)[0]),
	"tostring": lambda *args: toString(argumentValidation(args)[0]),
}