import bpy
import re
from ..tf_util.base_functions import TF_base_functions
from ..tf_util.mesh_util import TFMeshSetting

plugin_info = {
	"name": "Repaint Util",
	"author": "Merk",
	"version": "0.2.0",
}

class TFRepaintSetting(bpy.types.PropertyGroup):
	repaintable = bpy.props.BoolProperty(default=True)
	repaint_id = bpy.props.StringProperty(default="")

	@classmethod
	def register(cls):
		TFMeshSetting.repaint = bpy.props.PointerProperty(type=cls)
		bpy.types.Object.tf_repaint = bpy.props.PointerProperty(type=cls)

	@classmethod
	def unregister(cls):
		del TFMeshSetting.repaint
		del bpy.types.Object.tf_repaint

class TFRepaintUtil():
	def save_mesh_data_once(self, data):
		if TF_base_functions.user_prefs["debug"]:
			print("TFRepaintUtil save_mesh_data_once")
		
		if data["meshObj"]:
			if data["meshObj"].data.tf_mesh.repaint.repaintable:
				data["meshData"]["__repaintable"] = True
				data["meshData"]["__repaint_id"] = data["meshObj"].data.tf_mesh.repaint.repaint_id
	
	def save_group_data(self, data):
		if TF_base_functions.user_prefs["debug"]:
			print("TFRepaintUtil save_group_data")
		
		if data["groupObj"]:
			if data["groupObj"].tf_repaint.repaintable:
				data["groupData"]["__repaintable"] = True
				data["groupData"]["__repaint_id"] = data["groupObj"].tf_repaint.repaint_id

	def write_mesh_file(self, data):
		if TF_base_functions.user_prefs["debug"]:
			print("TFRepaintUtil write_mesh_file")
		
		if TF_base_functions.export_settings["repaintable"] and "__repaintable" in data["meshData"] and data["meshData"]["__repaintable"]:
			repaint_id = TF_base_functions.export_settings["repaint_id"]
			if "__repaint_id" in data["meshData"] and data["meshData"]["__repaint_id"] != "":
				repaint_id = data["meshData"]["__repaint_id"]
			
			if repaint_id != "":
				data["text"] = re.sub(
					r'function data\(\)\s*return(.*)end',
					r'local result = \1',
					data["text"],
					flags = re.DOTALL
				)
				data["text"] += '\nif merk_vehicleUtil and merk_vehicleUtil[1] then\n'
				data["text"] += '\tmerk_vehicleUtil[1].getMeshData("%s", "%s", result)\n' % (repaint_id, data["fileID"])
				data["text"] += 'end\n\nfunction data()\n\treturn result\nend'
	
	def write_group_file(self, data):
		if TF_base_functions.user_prefs["debug"]:
			print("TFRepaintUtil write_group_file")
		
		if TF_base_functions.export_settings["repaintable"] and "__repaintable" in data["groupData"] and data["groupData"]["__repaintable"]:
			repaint_id = TF_base_functions.export_settings["repaint_id"]
			if "__repaint_id" in data["groupData"] and data["groupData"]["__repaint_id"] != "":
				repaint_id = data["groupData"]["__repaint_id"]
				
			if repaint_id:
				data["text"] = re.sub(
					r'function data\(\)\s*return(.*)end',
					r'local result = \1',
					data["text"],
					flags = re.DOTALL
				)
				data["text"] += '\nif merk_vehicleUtil and merk_vehicleUtil[1] then\n'
				data["text"] += '\tmerk_vehicleUtil[1].getGroupData("%s", "%s", result)\n' % (repaint_id, data["fileID"])
				data["text"] += 'end\n\nfunction data()\n\treturn result\nend'
				
	def draw_mesh_settings_panel(self, data):
		if TF_base_functions.user_prefs["debug"]:
			print("TFRepaintUtil draw_mesh_settings_panel")

		boxRepaint = TF_base_functions.draw_hideable_box(data["layout"], "Repaints:", "mesh_repaint", True)
		if boxRepaint:
			boxRepaint.prop(data["mydata"].tf_mesh.repaint, "repaintable", text="Repaintable", toggle = True)
			boxRepaint.prop(data["mydata"].tf_mesh.repaint, "repaint_id", text="Repaint ID")
			data["layout"].separator()
				
	def draw_group_settings_panel(self, data):
		if TF_base_functions.user_prefs["debug"]:
			print("TFRepaintUtil draw_group_settings_panel")

		boxRepaint = TF_base_functions.draw_hideable_box(data["layout"], "Repaints:", "group_repaint", True)
		if boxRepaint:
			boxRepaint.prop(data["mydata"].tf_repaint, "repaintable", text="Repaintable", toggle = True)
			boxRepaint.prop(data["mydata"].tf_repaint, "repaint_id", text="Repaint ID")
			data["layout"].separator()

	def init_export_operator(self, data):
		if TF_base_functions.user_prefs["debug"]:
			print("TFRepaintUtil init_export_operator")
		
		if data["operator"]:
			if "repaintable" not in data["operator"].plugin_props:
				data["operator"].add_prop("repaintable", "bool", "Repaintable", "")
			if "repaint_id" not in data["operator"].plugin_props:
				data["operator"].add_prop("repaint_id", "string", "Repaint ID", "")
				
	def draw_export_settings_panel(self, data):
		if TF_base_functions.user_prefs["debug"]:
			print("TFRepaintUtil draw_export_settings_panel")
		
		if data["mydata"] and "repaintable" in data["mydata"].plugin_props and "repaint_id" in data["mydata"].plugin_props:
			data["layout"].separator()
			data["mydata"].plugin_props["repaintable"].draw(data["layout"], True)
			if data["mydata"].plugin_props["repaintable"].getValue():
				data["mydata"].plugin_props["repaint_id"].draw(data["layout"])


class_list = [
	TFRepaintSetting,
]

if "repaintable" not in TF_base_functions.user_settings["export_options"]:
	TF_base_functions.user_settings["export_options"]["repaintable"] = False
if "repaint_id" not in TF_base_functions.user_settings["export_options"]:
	TF_base_functions.user_settings["export_options"]["repaint_id"] = ""

def register():
	for class_type in class_list:
		bpy.utils.register_class(class_type)

	repaintUtil = TFRepaintUtil()
	TF_base_functions.plugin_add("TFRepaintUtil", repaintUtil)

def unregister():
	TF_base_functions.plugin_remove("TFRepaintUtil")

	for i in range(len(class_list)-1, -1, -1):
		bpy.utils.unregister_class(class_list[i])