import bpy
import re
from mathutils import *
from math import *
from ..tf_util.base_functions import TF_base_functions, TFPropertyGroup, LuaExportFunctionProto

plugin_info = {
	"name": "Transport Network Provider",
	"author": "eis_os, Merk",
	"version": "0.7.0",
	"order": 0,
}

transportModes = [
	["PERSON", "PERSON", "", 1],
	["CARGO", "CARGO", "", 2],
	["BUS", "BUS", "", 4],
	["TRUCK", "TRUCK", "", 8],
	["TRAM", "TRAM", "", 16],
	["ELECTRIC_TRAM", "ELECTRIC_TRAM", "", 32],
	["SHIP", "SHIP", "", 64],
	["AIRCRAFT", "AIRCRAFT", "", 128],
]

runwayTypes = [
	["NONE", "NONE", "", 0],
	["TAKEOFF", "TAKEOFF", "", 1],
	["LANDING", "LANDING", "", 2],
]

runwayTypeList = []
for runwayType in runwayTypes:
	runwayTypeList.append(runwayType[0])

class TFTransportNetworkGroupSettings(bpy.types.PropertyGroup):
	@classmethod
	def register(cls):
		bpy.types.Object.transportNetwork = bpy.props.PointerProperty(
						attr="transportNetwork",
						name="TF Transport Network Settings",
						description = "Transport network properties used by TF",
						type = cls,
						)
						
		cls.transportModes = bpy.props.EnumProperty(
			options = {'ENUM_FLAG'}, 
			default=set(),
			items=[(transportModes[i][1], transportModes[i][0], transportModes[i][2], transportModes[i][3]) for i in range(0, len(transportModes))]
			)
		cls.speedLimit = bpy.props.IntProperty(default = 20)
		
	@classmethod
	def unregister(cls):
		del bpy.types.Object.transportNetwork

class TFTransportNetworkCurveSettings(bpy.types.PropertyGroup):
	@classmethod
	def register(cls):
		if TF_base_functions.user_prefs["debug"]:
			print('TF Transport Network Curve Settings register')
		bpy.types.Curve.transportNetwork = bpy.props.PointerProperty(
						attr="transportNetwork",
						name='TF Transport Network Settings',
						description = 'Transport network properties used by TF',
						type = cls,
						)
		cls.personEdgeGroup =  bpy.props.IntProperty(name="personEdgeGroup", default=-1)
		cls.vehicleNodeGroupA = bpy.props.IntProperty(name="vehicleNodeGroupA", default=-1)
		cls.personNodeGroupA =  bpy.props.IntProperty(name="personNodeGroupA", default=-1)
		cls.vehicleNodeGroupB = bpy.props.IntProperty(name="vehicleNodeGroupB", default=-1)
		cls.personNodeGroupB =  bpy.props.IntProperty(name="personNodeGroupB", default=-1)
		cls.runwayTypeA = bpy.props.EnumProperty(
			name="runwayTypeA", default="NONE",
			items=[(runwayTypes[i][1], runwayTypes[i][0], runwayTypes[i][2], runwayTypes[i][3]) for i in range(0, len(runwayTypes))]
		)
		cls.runwayTypeB = bpy.props.EnumProperty(
			name="runwayTypeB", default="NONE",
			items=[(runwayTypes[i][1], runwayTypes[i][0], runwayTypes[i][2], runwayTypes[i][3]) for i in range(0, len(runwayTypes))]
		)
		cls.trackWidth = bpy.props.IntProperty(name="trackWidth", default=-1)
		
	@classmethod
	def unregister(cls):
		del bpy.types.Curve.transportNetwork

class TFTransportNetworkCurveSettingsAddOp(bpy.types.Operator):
	bl_label = "Add tf curve metadata"
	bl_idname = "collection.tfcurvedata_add"
	
	def invoke(self, context, event):
		for obj in context.selected_objects:
			if obj.type in {'CURVE', }:
				dummy = obj.data.transportNetwork
				
		return {'FINISHED'}

class TFTransportNetworkCurveSettingsRemoveOp(bpy.types.Operator):
	bl_label = "Remove tf curve metadata"
	bl_idname = "collection.tfcurvedata_remove"
	
	def invoke(self, context, event):
		del context.object.data['transportNetwork']
		return {'FINISHED'}
		
class TFTransportNetworkSettingsPanelOld(bpy.types.Panel):     
	bl_label = "Train Fever TransportNetworkSettings"
	bl_space_type = "PROPERTIES"
	bl_region_type = "WINDOW"
	bl_context = "data"
	
	@classmethod
	def poll(cls, context):
		return (context
				and context.object
				and context.object.type in {'CURVE', }
				and context.object.data
				and context.object.data.is_property_set('transportNetwork')
				)

	def draw(self, context):
		if context.object.type != 'CURVE':
			return;
		obj = context.curve
		if context.object.data.is_property_set('transportNetwork') == False:
			return;
		#mydata = obj.transportNetwork
		mydata = context.object.data.transportNetwork
		layout = self.layout
		box = layout.box()
		row = box.row()
		
		row = layout.row()
		row.label("transportModes:")
		boxModes = row.box();
		boxModes.prop(mydata, "transportModes",text="transportModes")
		box.prop(mydata, "personEdgeGroup",text="personEdgeGroup")
		box.prop(mydata, "vehicleNodeGroupA",text="vehicleNodeGroupA")
		box.prop(mydata, "vehicleNodeGroupB",text="vehicleNodeGroupB")
		box.prop(mydata, "personNodeGroupA",text="personNodeGroupA")
		box.prop(mydata, "personNodeGroupB",text="personNodeGroupB")
		box.prop(mydata, "trackWidth",text="Track width")
		
class TFTransportNetworkSettingsPanel(bpy.types.Panel):     
	bl_label = "Train Fever TransportNetwork"
	bl_space_type = "PROPERTIES"
	bl_region_type = "WINDOW"
	bl_context = "object"
	
	@classmethod
	def poll(cls, context):
		if (context and context.object and context.object.type in {'CURVE', 'EMPTY', }):
			if context.object.type == 'CURVE':
				return True
			elif context.object.type == 'EMPTY':
				if context.object.subtype in {'transportNetworkGroup', }:
					return True                
		return False

	def draw(self, context):      
		layout = self.layout
		box = layout.box()
		if context.object.type == 'EMPTY':
			subtype = context.object.subtype
			if subtype == 'transportNetworkGroup':
				self.draw_group(context, box)
		elif context.object.type == 'CURVE': 
			self.draw_curve(context, box)
			
	def draw_curve(self, context, box):
		if context.object.data.is_property_set('transportNetwork'):
			mydata = context.object.data.transportNetwork
			box.prop(mydata, "trackWidth",text="Track width")
			box.prop(mydata, "vehicleNodeGroupA",text="vehicleNodeGroupA")
			box.prop(mydata, "vehicleNodeGroupB",text="vehicleNodeGroupB")
			box.prop(mydata, "personNodeGroupA",text="personNodeGroupA")
			box.prop(mydata, "personNodeGroupB",text="personNodeGroupB")
			box.prop(mydata, "personEdgeGroup",text="personEdgeGroup")
			box.prop(mydata, "runwayTypeA",text="runwayTypeA")
			box.prop(mydata, "runwayTypeB",text="runwayTypeB")
			box.operator("collection.tfcurvedata_remove", icon="ZOOMOUT", text="")
		else:
			box.operator("collection.tfcurvedata_add", icon="ZOOMIN", text="")
	   
		
	def draw_group(self, context, box):
		mydata = context.object.transportNetwork
		layout = self.layout
		row = box.row()
		row.label("transportModes:")
		boxModes = row.box()
		boxModes.prop(mydata, "transportModes", text="transportModes")
		layout.prop(mydata, "speedLimit", text="speedLimit")
		
class TFTransportNetworkImportExport():
	def load_model(self, data):
		if TF_base_functions.user_prefs["debug"]:
			print("TFTransportNetworkImport load_model")
			
		model_data = data["modelData"]
		model = data["modelObj"]
		
		if "metadata" in model_data and "transportNetworkProvider" in model_data["metadata"]:
			if "laneLists" in model_data["metadata"]["transportNetworkProvider"]:    
				laneLists = self.read_transportNetworkProvider(model_data["metadata"]["transportNetworkProvider"])
				laneLists.parent = model
			
			if model_data["metadata"]["transportNetworkProvider"] == {}:
				del model_data["metadata"]["transportNetworkProvider"]
	
	def init_import_operator(self, data):
		if data["operator"]:
			if "path_layer" not in data["operator"].plugin_props:
				data["operator"].add_prop("path_layer", "bool", "Path Layer", "")
				
	def draw_import_settings_panel(self, data):
		if data["mydata"] and "path_layer" in data["mydata"].plugin_props:
			data["layout"].separator()
			data["mydata"].plugin_props["path_layer"].draw(data["layout"], True)
	
	def read_transportNetworkProvider(self, meta_network):
		laneLists = []
		if "laneLists" in meta_network:
			laneLists = meta_network["laneLists"]
			del meta_network["laneLists"]
		
		terminals = {}
		if "terminals" in meta_network:
			terminals = meta_network["terminals"]
			del meta_network["terminals"]

		transportNetworkProvider = bpy.data.objects.new("transportNetworkProvider", None)
		transportNetworkProvider.subtype = "transportNetworkProvider"
		
		bpy.context.scene.objects.link(transportNetworkProvider)

		pathLayer = TF_base_functions.import_settings["path_layer"]
		if pathLayer:
			maxLayer = 0
			for i in range(20):
				if bpy.context.scene.layers[i]:
					maxLayer = i+1
			
			if maxLayer > 19:
				maxLayer = 19
				
			transportNetworkProvider.layers[maxLayer] = True
			for i in range(20):
				transportNetworkProvider.layers[i] = (i == maxLayer)
		
		vehicleNode = {}
		personNodes = {}
		personEdges = {}
		runways = {}
		
		for i in range(len(terminals)):
			terminalEntry = terminals[i]
			if "personEdges" in terminalEntry:
				for personEdgeNum in terminalEntry["personEdges"]:
					personEdges[personEdgeNum] = i
			if "personNodes" in terminalEntry:
				for personNodeNum in terminalEntry["personNodes"]:
					personNodes[personNodeNum] = i
			if "vehicleNode" in terminalEntry:
				vehicleNode[terminalEntry["vehicleNode"]] = i;
		
		if "runways" in meta_network:
			for i in range(len(meta_network["runways"])):
				runway = meta_network["runways"][i]
				if isinstance(runway, dict) and "node" in runway and "type" in runway and runway["type"] in runwayTypeList:
					runways[runway["node"]] = runway["type"]
					
			del meta_network["runways"]
				
		curEdge = 0
		curNode = 0
		for i in range(len(laneLists)):
			laneList = laneLists[i]
			transportModes = laneList["transportModes"]
			if "transportModes" in laneList:
				lanelistname = "_".join(laneList["transportModes"])
			else:
				lanelistname = "lane_"+str(i);
			
			transportNetworkGroup = bpy.data.objects.new(lanelistname, None)
			transportNetworkGroup.subtype = "transportNetworkGroup"
			transportNetworkGroup.transportNetwork.transportModes = set(transportModes)
			if "speedLimit" in laneList:
				transportNetworkGroup.transportNetwork.speedLimit = laneList["speedLimit"]
			
			bpy.context.scene.objects.link(transportNetworkGroup)
			if pathLayer:			
				transportNetworkGroup.layers[maxLayer] = True
				for j in range(20):
					transportNetworkGroup.layers[j] = (j == maxLayer)
			
			transportNetworkGroup.parent = transportNetworkProvider
			
			if "nodes" in laneList:
				for index in range(0, len(laneList["nodes"]), 2):
					lanename = lanelistname+"_"+str(index)
					
					nodeData1 = laneList["nodes"][index];
					nodeData2 = laneList["nodes"][index+1];
					
					curvedata = bpy.data.curves.new(name = lanename, type='CURVE')  
					curvedata.dimensions = '3D'
					objectdata = bpy.data.objects.new(lanename, curvedata)
					bpy.context.scene.objects.link(objectdata)
					objectdata.parent = transportNetworkGroup
					if pathLayer:					
						objectdata.layers[maxLayer] = True
						for j in range(20):
							objectdata.layers[j] = (j == maxLayer)
					
					if curEdge in personEdges:
#                        print ('Found personEdges @ '+lanename+' '+str(curEdge))
						curvedata.transportNetwork.personEdgeGroup = personEdges[curEdge]
						
					if curNode in personNodes:
#                        print ('Found personNodes @ '+lanename+' '+str(curNode))
						curvedata.transportNetwork.personNodeGroupA = personNodes[curNode]
					if curNode in vehicleNode:
						curvedata.transportNetwork.vehicleNodeGroupA = vehicleNode[curNode]
					if curNode in runways:
						curvedata.transportNetwork.runwayTypeA = runways[curNode]
					
					curNode +=1
					
					if curNode in personNodes:
#                        print ('Found personNodes @ '+lanename+' '+str(curNode))
						curvedata.transportNetwork.personNodeGroupB = personNodes[curNode]
					if curNode in vehicleNode:
						curvedata.transportNetwork.vehicleNodeGroupB = vehicleNode[curNode]
					if curNode in runways:
						curvedata.transportNetwork.runwayTypeB = runways[curNode]
					
					curNode +=1
					
					curvedata.transportNetwork.trackWidth = nodeData1[2]
					
					polyline = curvedata.splines.new('BEZIER')
					points = polyline.bezier_points
					points.add(1)
					self.read_laneLists_createpoint(points[0], nodeData1, 0)
					self.read_laneLists_createpoint(points[1], nodeData2, 1)
					
					curEdge += 1
					
		return transportNetworkProvider
   
	def read_laneLists_createpoint(self, point, nodeData, side):
		vec1 = Vector(nodeData[0])
		vec2 = Vector(nodeData[1])
		
		vec2half = vec2*0.5
		vec3 = vec1+vec2half
		vec4 = vec1-vec2half
		
		width = nodeData[2]
		point.co = vec1
		
		if (side == 0):
			point.handle_left_type = 'ALIGNED'
			point.handle_right_type = 'FREE'
		
			point.handle_left = vec4
			point.handle_right = vec3
		else:
			point.handle_left_type = 'FREE'
			point.handle_right_type = 'ALIGNED'
		
			point.handle_left = vec4
			point.handle_right = vec3
		
		
	def read_laneLists_createpoint_old(self, point, nodeData):
		vec1 = Vector(nodeData[0])
		vec2 = Vector(nodeData[1])
		vec3 = vec2+vec2

		width = nodeData[2]
		point.co = (vec1[0], vec1[1], vec1[2])
	
		point.handle_left_type = 'VECTOR'
		point.handle_right_type = 'VECTOR'
		point.handle_right = vec3     
		
	def laneUtilsMakeLane(self, ep0, ep1, et0, et1, w):
		p0 = Vector(ep0[0], ep0[1], ep0[2])
		p1 = Vector(ep1[0], ep1[1], ep1[2])
		
		t0 = Vector(et0[0], et0[1], et0[2]).normalized()
		t1 = Vector(et0[0], et0[1], et0[2]).normalized()
	
		length = (p0 - p1).length
		angle = vec1.angle(vec2)
		
		scale = calcScale(length, angle)

		t0 *= scale
		t1 *= scale

	def save_model_data(self, data):
		if TF_base_functions.user_prefs["debug"]:
			print("TFTransportNetworkExport save_model_data")
		
		model_data = data["modelData"]
		main_object = data["modelObj"]
		
		if "metadata" not in model_data:
			model_data["metadata"] = {}
		
		for obj in main_object.children:
			if obj.type == 'EMPTY':
				if obj.subtype == "transportNetworkProvider":
					transportNetworkProvider = self.export_transportNetwork(obj)
					model_data["metadata"]["transportNetworkProvider"] = transportNetworkProvider
					if "__requires" not in model_data:
						model_data["__requires"] = []
					model_data["__requires"].append("local laneutil = require \"laneutil\"")
		
		if model_data["metadata"] == {}:
			del model_data["metadata"]
				
	
	def natsortObjects(self, list_):
		# decorate
		tmp = []
		i = len(list_)
		for o in list_:
			m = re.search('\d+', getattr(o, "name", "") )
			if m:
				tmp.append((int(m.group(0)), o))
			else:
				tmp.append((i, o))
				i += 1
			 
		tmp.sort()
		# undecorate
		return [ i[1] for i in tmp ]
	
	def sortNetworkGroups(self, groups):
		sorted_groups = []
		
		for groupObj in groups:
			if groupObj.type == 'EMPTY':
				if groupObj.subtype == 'transportNetworkGroup':
					if "SHIP" in groupObj.transportNetwork.transportModes:
						sorted_groups = [groupObj] + sorted_groups
					else:
						sorted_groups.append(groupObj)
		
		return sorted_groups
	
	def export_transportNetwork(self, tNP):
		terminals = {}
		runways = []
		laneLists = []
		nodeId = 0
		
		networkGroups = self.sortNetworkGroups(tNP.children)
		for groupObj in networkGroups:
			nodes = []
			groupObjList = self.natsortObjects( groupObj.children )
			for curveObj in groupObjList:
				if curveObj.type == 'CURVE':
					if curveObj.data.is_property_set('transportNetwork'):
						transportNetwork = curveObj.data.transportNetwork
						trackWidth = transportNetwork.trackWidth
						
						self.insertTerminalData(terminals, transportNetwork, nodeId)
						self.insertRunwayData(runways, transportNetwork, nodeId)
						
						points = curveObj.data.splines[0].bezier_points
						for i in range(1, len(points)):
							nodes.extend( self.createNodesLaneUtils(points[i-1], points[i], trackWidth, curveObj.location) )
							nodeId+=2

			group = {
				"transportModes": list(groupObj.transportNetwork.transportModes),
				"speedLimit": groupObj.transportNetwork.speedLimit,
				"nodes": LuaExportFunctionProto("laneutil.makeLanes", [nodes])
			}
			
			laneLists.append(group)
					
		transportNetworkProvider = {
			"laneLists": laneLists,
			"terminals": list(terminals.values())
		}
		
		if len(terminals.values()) > 0:
			transportNetworkProvider["terminals"] = list(terminals.values())
		
		if len(runways) > 0:
			transportNetworkProvider["runways"] = runways
		
		return transportNetworkProvider
	   
	def createNodes(self, pointA, pointB, width, location):
		p0 = pointA.co
		t0 = (pointA.handle_right-p0)*2
		
		p1 = pointB.co
		t1 = (pointB.handle_right-p1)*2
		
		p0 = p0 + location
		p1 = p1 + location
		
		nodes = [
			[list(p0), list(t0), width],
			[list(p1), list(t1), width]
		]
		return nodes
		
	def createNodesLaneUtils(self, pointA, pointB, width, location):
		p0 = pointA.co
		t0 = (pointA.handle_right-p0)*2
		
		p1 = pointB.co
		t1 = (pointB.handle_right-p1)*2
		
		p0 = p0 + location
		p1 = p1 + location
		
		nodes = [
			[list(p0), list(p1), list(t0), list(t1), width],
		]
		return nodes
	
	def insertTerminalData(self, terminals, transportNetwork, nodeId):
		mapping = {
			'personNodeGroupA': ['personNodes', 0, 1],
			'personNodeGroupB': ['personNodes', 1, 1],
			'vehicleNodeGroupA': ['vehicleNode', 0, 1],
			'vehicleNodeGroupB': ['vehicleNode', 1, 1],
			'personEdgeGroup': ['personEdges', 0, 2]
		}
		
		# create terminal to nodeID / edgeID Mapping
		for (blenderkey, tfmap) in mapping.items():
			terminalId = getattr(transportNetwork, blenderkey)
			if (terminalId != -1):
				nodeEdgeIdAdjusted = (nodeId + tfmap[1]) // tfmap[2]
				self.setTerminalAttribute(terminals, terminalId, tfmap[0], nodeEdgeIdAdjusted)

	def setTerminalAttribute(self, terminals, terminalId, option, val):
		if terminalId not in terminals:
			terminals[terminalId] = {}
			
		if option in {'personNodes', 'personEdges'}:
			if option not in terminals[terminalId]:
				terminals[terminalId][option] = []
			terminals[terminalId][option].append(val)
		elif option == 'vehicleNode':
			terminals[terminalId][option] = val
		else:
			showError('Unknown setTerminalAttribute')
	
	def insertRunwayData(self, runways, transportNetwork, nodeId):
		mapping = {
			'runwayTypeA': [0, 2],
			'runwayTypeB': [1, 2],
		}
		
		# create runway to nodeID / edgeID Mapping
		for (blenderkey, tfmap) in mapping.items():
			runwayType = getattr(transportNetwork, blenderkey)
			if runwayType != "NONE":
				runways.append({
					"type": runwayType,
					"node": nodeId + tfmap[0],
					"edges": [(nodeId + tfmap[0]) // tfmap[1]],
				})
				

if "path_layer" not in TF_base_functions.user_settings["import_options"]:
	TF_base_functions.user_settings["import_options"]["path_layer"] = False

def register():
	bpy.utils.register_class(TFTransportNetworkGroupSettings)
	
	bpy.utils.register_class(TFTransportNetworkCurveSettingsAddOp)
	bpy.utils.register_class(TFTransportNetworkCurveSettingsRemoveOp)
	bpy.utils.register_class(TFTransportNetworkCurveSettings) 
	
	bpy.utils.register_class(TFTransportNetworkSettingsPanel)

	TF_base_functions.subtype_list.append(('transportNetworkProvider', 'Transport Network Provider', '', 512))
	TF_base_functions.subtype_list.append(('transportNetworkGroup', 'Transport Network Group', '', 514))
	TF_base_functions.plugin_add("TFTransportNetworkImportExport", TFTransportNetworkImportExport())
	
def unregister():
	TF_base_functions.plugin_remove("TFTransportNetworkImportExport")
	TF_base_functions.subtype_list.remove(('transportNetworkProvider', 'Transport Network Provider', '', 512))
	TF_base_functions.subtype_list.remove(('transportNetworkGroup', 'Transport Network Group', '', 514))
		
	bpy.utils.unregister_class(TFTransportNetworkSettingsPanel)
		
	bpy.utils.unregister_class(TFTransportNetworkCurveSettings)
	bpy.utils.unregister_class(TFTransportNetworkCurveSettingsAddOp)
	bpy.utils.unregister_class(TFTransportNetworkCurveSettingsRemoveOp)
	
	bpy.utils.unregister_class(TFTransportNetworkGroupSettings)