import bpy
from ..lua import writer
from . import mesh_util, group_util, UI_util
from .base_functions import TF_base_functions
from mathutils import Vector

# Import
def is_valid_import_type(file_ext):
	return file_ext == "mdl"

def validate_children_TPF2(children, childrenName, fileName, errors):
	if len(children) > 0:
		for i in range (0, len(children)):
			child = children[i]
			if str(type(child)) == "<class 'dict'>":
				#meshRequired = True
				if "children" in child: # Recurse
					validate_children_TPF2(children[i]["children"], "%s[%i].children" % (childrenName, i), fileName, errors)
					#meshRequired = False
				
				#if meshRequired and "mesh" not in child: # Fairlie has a child without a mesh
					#errors.append({"level": 3, "message": "Missing field \"%s[%i].mesh\" in Model \"%s\"" % (childrenName, i, fileName)})
					#return False
				
				if "mesh" in child:
					if str(type(child["mesh"])) != "<class 'str'>":
						errors.append({"level": 3, "message": "Wrong field \"%s[%i].mesh\" in Model \"%s\"" % (childrenName, i, fileName)})
						return False
					if "materials" not in child:
						errors.append({"level": 3, "message": "Missing field \"%s[%i].materials\" in Model \"%s\"" % (childrenName, i, fileName)})
						return False 
					if str(type(child["materials"])) != "<class 'list'>":
						errors.append({"level": 3, "message": "Wrong field \"%s[%i].materials\" in Model \"%s\"" % (childrenName, i, fileName)})
						return False 
					if "transf" not in child :
						errors.append({"level": 3, "message": "Missing field \"%s[%i].transf\" in Model \"%s\"" % (childrenName, i, fileName)})
						return False 
					if str(type(child["transf"])) != "<class 'list'>" or len(child["transf"]) < 16:
						errors.append({"level": 3, "message": "Wrong field \"%s[%i].transf\" in Model \"%s\"" % (childrenName, i, fileName)})
						return False
			else:
				errors.append({"level": 3, "message": "Field \"%s[%i]\" in Model \"%s\" has a wrong data format!" % (childrenName, i, fileName)})
				return False
	else:
		errors.append({"level": 3, "message": "Field \"%s\" in Model \"%s\" has no elements!" % (childrenName, fileName)})
		
	return True

def _validate_data(data, errors):
	TF_base_functions.call_plugin("validate_model_data", data, errors)

	if data["modelData"]:
		# check lods
		if "lods" in data["modelData"] and str(type(data["modelData"]["lods"])) == "<class 'list'>":
			if len(data["modelData"]["lods"]) > 0:
				for i in range(0, len(data["modelData"]["lods"])):
					lod = data["modelData"]["lods"][i]
					if str(type(lod)) == "<class 'dict'>":
						if "static" not in lod or str(type(lod["static"])) != "<class 'bool'>":
							errors.append({"level": 3, "message": "Missing or wrong field \"lods[%i].static\" in Model \"%s\"!" % (i, data["fileName"])})
							return False
						if "visibleFrom" not in lod or str(type(lod["visibleFrom"])) not in ("<class 'int'>", "<class 'float'>", "<class 'long'>"):
							errors.append({"level": 3, "message": "Missing or wrong field \"lods[%i].visibleFrom\" in Model \"%s\"!" % (i, data["fileName"])})
							return False
						if "visibleTo" not in lod or str(type(lod["visibleTo"])) not in ("<class 'int'>", "<class 'float'>", "<class 'long'>"):
							errors.append({"level": 3, "message": "Missing or wrong field \"lods[%i].visibleTo\" in Model \"%s\"!" % (i, data["fileName"])})
							return False

						# Children
						if TF_base_functions.import_settings["game"] == "TPF2":
							if "node" in lod and str(type(lod["node"])) == "<class 'dict'>":
								if "children" in lod["node"] and str(type(lod["node"]["children"])) == "<class 'list'>":
									if not validate_children_TPF2(lod["node"]["children"], "lods[%i].node.children" % i, data["fileName"], errors):
										return False
								else:
									errors.append({"level": 3, "message": "Missing or wrong field \"lods[%i].node.children\" in Model \"%s\"!" % (i, data["fileName"])})
									return False
							else:
								errors.append({"level": 3, "message": "Missing or wrong field \"lods[%i].node\" in Model \"%s\"!" % (i, data["fileName"])})
								errors.append({"level": 3, "message": "Type: %s" % str(type(lod["node"]))})
								return False
						else:
							if "children" in lod and str(type(lod["children"])) == "<class 'list'>":
								if len(lod["children"]) > 0:
									for j in range(0, len(lod["children"])):
										child = lod["children"][j]
										if str(type(child)) == "<class 'dict'>":
											if "id" not in child or str(type(child["id"])) != "<class 'str'>":
												errors.append({"level": 3, "message": "Missing or wrong field \"lods[%i].children[%i].id\" in Model \"%s\"!" % (i, j, data["fileName"])})
												return False
											if "type" not in child or str(type(child["type"])) != "<class 'str'>":
												errors.append({"level": 3, "message": "Missing or wrong field \"lods[%i].children[%i].type\" in Model \"%s\"!" % (i, j, data["fileName"])})
												return False
											if "transf" not in child or str(type(child["transf"])) != "<class 'list'>" or len(child["transf"]) < 16:
												errors.append({"level": 3, "message": "Missing or wrong field \"lods[%i].children[%i].transf\" in Model \"%s\"!" % (i, j, data["fileName"])})
												return False
										else:
											errors.append({"level": 3, "message": "Field \"lods[%i].children[%i]\" in Model \"%s\" has a wrong data format!" % (i, j, data["fileName"])})
											return False
								else:
									errors.append({"level": 2, "message": "Lod %i in Model \"%s\" has no children!" % (i, data["fileName"])})
							else:
								errors.append({"level": 3, "message": "Missing or wrong field \"lods[%i].children\" in Model \"%s\"!" % (i, data["fileName"])})
								return False

						if TF_base_functions.import_settings["game"] != "TPF2":
							if "matConfigs" not in lod or str(type(lod["matConfigs"])) != "<class 'list'>":
								errors.append({"level": 3, "message": "Missing or wrong field \"lods[%i].matConfig\" in Model \"%s\"!" % (i, data["fileName"])})
								return False
							else:
								if str(type(lod["matConfigs"][0])) == "<class 'list'>":
									if len(lod["children"]) > len(lod["matConfigs"][0]):
										errors.append({"level": 2, "message": "MatConfig of Lod %i in Model \"%s\" has not enough values!" % (i, data["fileName"])})
										for k in range(len(lod["matConfigs"][0]), len(lod["children"])):
											lod["matConfigs"][0].append(0)
									elif len(lod["children"]) < len(lod["matConfigs"][0]):
										errors.append({"level": 1, "message": "MatConfig of Lod %i in Model \"%s\" has too much values!" % (i, data["fileName"])})
								else:
									errors.append({"level": 3, "message": "Field \"lods[%i].matConfigs[%i]\" in Model \"%s\" has a wrong data format!" % (i, j, data["fileName"])})
									return False
					else:
						errors.append({"level": 3, "message": "Field \"lods[%i]\" in Model \"%s\" has a wrong data format!" % (i, data["fileName"])})
						return False
			else:
				errors.append({"level": 2, "message": "Model \"%s\" has no lods!" % data["fileName"]})
		else:
			errors.append({"level": 3, "message": "Missing or wrong field \"lods\" in Model \"%s\"!" % data["name"]})
			return False
	else:
		errors.append({"level": 3, "message": "Model \"%s\" could not be loaded or has a wrong data format!" % data["name"]})
		return False

	return True

def load_children_TPF2(basepath, children, parent, id_to_object, parentName, fileName, select_all, errors):
	if len(children) > 0:
		for i in range (0, len(children)):
			child = children[i]
			child_obj = None
			if "mesh" in child:
				childName = ""
				if "name" in child:
					childName = child["name"]
					
				# if (TF_base_functions.user_prefs["debug"]):
					# print("\tKeys in model %s" % childName)
					# for s in child:
						# print("\t\t%s" % s)
					
				child_filename = child["mesh"]
				child_matlist = child["materials"]
				child_obj = mesh_util.load_from_file_TPF2(basepath + "models/mesh/" + child_filename, childName, child_matlist, select_all, errors)
				child_obj.index = i
				id_to_object += [(child_obj, child_filename)]
				if child_obj:
					child_obj.matrix_local = TF_base_functions.create_4x4matrix(child['transf'])
					child["blender_object"] = child_obj
					child_obj.parent = parent
					data = {
						"modelData": child,
						"fileName": fileName,
						"modelObj": child_obj,
						"name": childName,
						"basepath": basepath,
					}
					TF_base_functions.call_plugin("load_model_2", data, errors)
					
			if "children" in child: # Recurse
				childName = ""
				if "name" in child:
					childName = child["name"]
				else:
					childName = parentName+"["+str(i)+"]"
					
				new_group = child_obj
				if not new_group:
					new_group = bpy.data.objects.new(childName, None)
					new_group.index = i
					new_group.subtype = "group"
					new_group.model_type = parent.model_type
					new_group.parent = parent
					new_group.select = parent.select
					new_group.matrix_local = TF_base_functions.create_4x4matrix(child['transf'])
					
					scene = bpy.context.scene
					scene.objects.link(new_group)
				
				load_children_TPF2(basepath, children[i]["children"], new_group, id_to_object, childName, fileName, select_all, errors)
	else:
		errors.append({"level": 3, "message": "Field \"%s\" in Model \"%s\" has no elements!" % (childrenName, fileName)})

def load_from_file(filename, matConfig, select, errors, select_all = False):
	if (TF_base_functions.user_prefs["debug"]):
		print("load_model_file", filename)

	lod_distance = TF_base_functions.user_settings["import_options"]["lod_distance"]
	if "lod_distance" in TF_base_functions.import_settings:
		lod_distance = TF_base_functions.import_settings["lod_distance"]

	basepath, filepath, model_name = TF_base_functions.split_path(filename, errors)
	if basepath == "":
		errors.append({"level": 1, "message": "Could not find \"res\"-folder in \"%s\", loading of childrens probably won't be succesfull." %filename})
		basepath = TF_base_functions.import_settings["basepath"] + "/res/"

	model_name = model_name.replace(".mdl", "")

	model_data = TF_base_functions.get_data(basepath + filepath, "MODEL", errors)
	data = {
		"modelData": model_data,
		"fileName": filename,
		"name": model_name,
		"basepath": basepath,
	}

	if _validate_data(data, errors):
		new_model = bpy.data.objects.new(model_name, None)
		new_model.subtype = 'model'
		new_model.model_type = TF_base_functions.get_model_type(filename, "models/model/")
		new_model.select = select

		scene = bpy.context.scene
		scene.objects.link(new_model)
		if select:
			scene.objects.active = new_model

		if "boundingInfo" in model_data and str(type(model_data["boundingInfo"])) == "<class 'dict'>":
			if "bbMax" in model_data["boundingInfo"]:
				new_model.tf_boundingInfo.bbMax = model_data["boundingInfo"]["bbMax"]
			if "bbMin" in model_data["boundingInfo"]:
				new_model.tf_boundingInfo.bbMin = model_data["boundingInfo"]["bbMin"]
				
		data["modelObj"] = new_model
		TF_base_functions.call_plugin("load_model_data", data, errors)

		for i in range(0, len(model_data["lods"])):
			if (lod_distance == 0) or ((model_data["lods"][i]["visibleTo"] >= lod_distance) and (model_data["lods"][i]["visibleFrom"] <= lod_distance)):
				# create lod and set data
				new_group = bpy.data.objects.new(model_name + "_lod_" + str(i), None)
				new_group.subtype = "lod"
				new_group.tf_lod.static = model_data["lods"][i]["static"]
				new_group.tf_lod.visibleTo = model_data["lods"][i]["visibleTo"]
				new_group.tf_lod.visibleFrom = model_data["lods"][i]["visibleFrom"]
				scene = bpy.context.scene
				new_group.parent = new_model
				new_group.select = select_all
				scene.objects.link(new_group)

				# load children for lod
				id_to_object = []
				if TF_base_functions.import_settings["game"] == "TPF2":
					nodeName = ""
					if "name" in model_data["lods"][i]["node"]:
						nodeName = model_data["lods"][i]["node"]["name"]
					else:
						nodeName = model_name + "_lod_" + str(i) + "_RootNode"
						
					rootNode = bpy.data.objects.new(nodeName, None)
					rootNode.subtype = "node"
					rootNode.parent = new_group
					rootNode.select = select_all
					scene.objects.link(rootNode)
						
					load_children_TPF2(basepath, model_data["lods"][i]["node"]["children"], rootNode, id_to_object, nodeName, filename, select_all, errors)
				else:
					j = 0
					for child in model_data["lods"][i]["children"]:
						child_filename = child["id"]
						child_obj = None
						if child["type"] == 'MESH':
							child_obj = mesh_util.load_from_file(basepath + "models/mesh/" + child_filename, model_data["lods"][i]["matConfigs"][0][j], select_all, errors)
							id_to_object += [(child_obj, child_filename)]
						elif child["type"] == 'GROUP':
							child_obj, ids = group_util.load_from_file(basepath + "models/group/" + child_filename, model_data["lods"][i]["matConfigs"][0][j], select_all, errors, select_all)
							id_to_object += [(child_obj, child_filename)] + ids
						else:
							id_to_object += [(child_obj, child_filename)]
						
						if child_obj:
							child_obj.matrix_local = TF_base_functions.create_4x4matrix(child['transf'])
							child["blender_object"] = child_obj
							if new_group:
								child_obj.parent = new_group
						j = j + 1
				
				model_data["lods"][i]["__idToObject"] = id_to_object

				data["lodData"] = model_data["lods"][i]
				data["lodObj"] = new_group
				data["lodIndex"] = i
				TF_base_functions.call_plugin("load_lod", data, errors)

		# delete lod data from data dict
		if "lodData" in data:
			del data["lodData"]
		if "lodObj" in data:
			del data["lodObj"]

		data["modelObj"] = new_model
		if TF_base_functions.import_settings["at_cursor"]:
			new_model.location = bpy.context.scene.cursor_location
		TF_base_functions.call_plugin("load_model", data, errors)

		return new_model

def ui_only_cleanup(context):
	#garbage collection, delete all created objects
	object = context.scene.objects.active
	if object:
		objects = TF_base_functions.get_children(object, True)
		context.scene.objects.unlink(object)
		bpy.data.objects.remove(object)
		for obj in objects:
			context.scene.objects.unlink(obj)
			obj_data = None
			obj_type = None
			if obj.data:
				obj_data = obj.data
				obj_type = obj.type

			bpy.data.objects.remove(obj)
			if obj_data and obj_data.users == 0:
				if obj_type == 'MESH':
					bpy.data.meshes.remove(obj_data)
				elif obj_type == 'ARMATURE':
					bpy.data.armatures.remove(obj_data)

# Export
def is_valid_object(object):
	return object.type in {'EMPTY', 'ARMATURE'} and object.subtype == "model"

def get_export_type(object):
	return "mdl"

def get_model_type(object, model_type, model_type_other):
	return TF_base_functions.get_modeltype(object.model_type, model_type, model_type_other)

def get_mesh_dir(object, mesh_dir):
	return object.name.partition(".")[0]

def _get_child_data_TPF2(child, model_obj, model_type, bounding_box, errors):
	
	result = {}
	
	# if (TF_base_functions.user_prefs["debug"]):
		# print("_get_child_data_TPF2 with child = %s" % child.name, bounding_box["bbMin"][0])
	
	if len(child.children) > 0:
		result["children"] = []
		childPrinted = [ False for i in range(len(child.children)) ]
		for i in range(len(child.children)): # The simplest way to print IN ORDER is to do two loops
			for j in range(len(child.children)):
				childChild = child.children[j]
				if childChild.index == i:
					result["children"].append(_get_child_data_TPF2(childChild, model_obj, model_type, bounding_box, errors))
					childPrinted[j] = True
					continue
				# If this point is reached, the index actually was not found
				
		# Print any additional entries which were skipped in the loop above. This would happen if two children shared the same index (which is not expected, but also not guarded against)
		for i in range(len(child.children)): 
			if not childPrinted[i]:
				childChild = child.children[j]
				result["children"].append(_get_child_data_TPF2(childChild, model_obj, model_type, bounding_box, errors))
			
	result["name"] = child.name
	result["transf"] = []
	
	for i in range(4):
		for j in range(4):
			result["transf"].append(round(child.matrix_local.col[i][j], 5))
	
	if child.type == "MESH":
		child_model_type = child.data.model_type
		if child_model_type == "%auto":
			child_model_type = TF_base_functions.export_settings["model_type"]
		elif child_model_type == "%inherit":
			child_model_type = model_type

		child_mesh_dir = child.data.mesh_dir
		if child_mesh_dir == "%auto":
			child_mesh_dir = TF_base_functions.export_settings["mesh_directory"]
		child_id = child_model_type + child_mesh_dir + "/"[:len(child_mesh_dir)] + TF_base_functions.export_settings["prefix"] + child.data.filename + ".msh"

		# get matConfig an create child data
		mesh_object = {
			"object": child,
			"parent": {"object": model_obj, "model_type": model_type}
		}
		materials, bound_box = mesh_util.create_file_data(mesh_object, "", errors)
		
		result["materials"] = materials
		result["mesh"] = child_id
		
		data = {
			"modelData": result,
			# "fileName": child.data.filename,
			"modelObj": child,
			# "name": child.name,
		}
		TF_base_functions.call_plugin("save_model_data_2", data, errors)
		
		if bounding_box and bound_box:
			for key in {"bbMin", "bbMax"}:
				for i in range(3):
					vec = child.matrix_world*Vector(bound_box[key])
					if vec[i] < bounding_box["bbMin"][i]:
						bounding_box["bbMin"][i] = vec[i]
					if vec[i] > bounding_box["bbMax"][i]:
						bounding_box["bbMax"][i] = vec[i]
	
	return result
	
def _get_lod_data_TPF2(lod_object, children, model_obj, model_type, bounding_box, errors):
	lod_data = {"node": [], "static": False, "visibleFrom": 0, "visibleTo": 2500}
	if lod_object:
		lod_data["static"] = lod_object.tf_lod.static
		lod_data["visibleFrom"] = lod_object.tf_lod.visibleFrom
		lod_data["visibleTo"] = lod_object.tf_lod.visibleTo
	
	renderObjects = []
	childObjects = TF_base_functions.get_children(children)
	if lod_data["visibleFrom"] == 0 and TF_base_functions.export_settings["ui_icon"]:
		renderObjects = TF_base_functions.get_children(children, False) # FIX LATER?
		
	# Need to collect the overall bounding box AND insert all of the data into lod_data
	
	child_data = _get_child_data_TPF2(children[0], model_obj, model_type, bounding_box, errors) # set bounding_box by reference
	
	lod_data["node"] = child_data

	data = {
		"lodData": lod_data,
		"lodObj": lod_object,
		"children": childObjects,
	}
	TF_base_functions.call_plugin("save_lod_data", data, errors)

	return lod_data, renderObjects

def _get_lod_data(lod_object, children, model_obj, model_type, bounding_box, errors):
	lod_data = {"children": [], "matConfigs": [[]], "static": False, "visibleFrom": 0, "visibleTo": 2500}
	if lod_object:
		lod_data["static"] = lod_object.tf_lod.static
		lod_data["visibleFrom"] = lod_object.tf_lod.visibleFrom
		lod_data["visibleTo"] = lod_object.tf_lod.visibleTo
	
	renderObjects = []
	childObjects = TF_base_functions.get_children(children)
	if lod_data["visibleFrom"] == 0 and TF_base_functions.export_settings["ui_icon"]:
		renderObjects = TF_base_functions.get_children(children, False)

	for child in children:
		matConfig_id = -1
		bound_box = None
		transf = []
		for i in range(4):
			for j in range(4):
				transf.append(child.matrix_local.col[i][j])

		child_name = ""
		child_type = ""
		child_id = ""

		if child.type == 'MESH':
			child_type = "MESH"
			child_name = child.data.name

			# get model_type and mesh dir for child
			child_model_type = child.data.model_type
			if child_model_type == "%auto":
				child_model_type = TF_base_functions.export_settings["model_type"]
			elif child_model_type == "%inherit":
				child_model_type = model_type

			child_mesh_dir = child.data.mesh_dir
			if child_mesh_dir == "%auto":
				child_mesh_dir = TF_base_functions.export_settings["mesh_directory"]
			child_id = child_model_type + child_mesh_dir + "/"[:len(child_mesh_dir)] + child_name + ".msh"

			# get matConfig an create child data
			mesh_object = {
				"object": child,
				"parent": {"object": model_obj, "model_type": model_type}
			}
			matConfig_id, bound_box = mesh_util.create_file_data(mesh_object, "", errors)

		elif child.type in {'EMPTY', 'ARMATURE'} and (child.subtype == "group" or child.subtype == "node"):
			if child.subtype == "group":
				child_type = "GROUP"
			else:
				child_type = "NODE"
			child_name = child.name.partition(".")[0]

			# get model_type for child
			child_model_type = child.model_type
			if child_model_type == "%auto":
				child_model_type = TF_base_functions.export_settings["model_type"]
			elif child_model_type == "%inherit":
				child_model_type = model_type
			child_id = child_model_type + child_name + ".grp"

			# get matConfig an create child data
			grp_object = {
				"object": child,
				"parent": {"object": model_obj, "model_type": model_type}
			}
			if TF_base_functions.export_settings["game"] == "TPF2":
				bound_box = group_util.create_file_data_TPF2(grp_object, "", errors)
			else:
				matConfig_id, bound_box = group_util.create_file_data(grp_object, "", errors)

		else:
			# invalid child
			errors.append({"level": 1, "message": "Invalid type for child \"%s\" of model \"%s\", this child will be ignored!" %(child.name, model_obj.name)})

		# BB calculation
		if bounding_box and bound_box:
			for key in {"bbMin", "bbMax"}:
				for i in range(3):
					vec = child.matrix_local*Vector(bound_box[key])
					if vec[i] < bounding_box["bbMin"][i]:
						bounding_box["bbMin"][i] = vec[i]
					if vec[i] > bounding_box["bbMax"][i]:
						bounding_box["bbMax"][i] = vec[i]

		if TF_base_functions.export_settings["game"] != "TPF2":
			if child_id and child_type:
				if matConfig_id == -1:
					matConfig_id = 0
				lod_data["children"].append({"id": child_id, "transf": transf, "type": child_type})
				lod_data["matConfigs"][0].append(matConfig_id)

	data = {
		"lodData": lod_data,
		"lodObj": lod_object,
		"children": childObjects,
	}
	TF_base_functions.call_plugin("save_lod_data", data, errors)

	return lod_data, renderObjects

def create_file_data(model_object, name, errors):
	model_obj = model_object["object"]

	# get name
	model_name = name
	if model_name == "":
		model_name = model_obj.name
	model_name = TF_base_functions.export_settings["prefix"] + model_name

	if (TF_base_functions.user_prefs["debug"]):
		print("save_model_file", model_name)

	if "models" not in TF_base_functions.export_settings["filedata"]:
		TF_base_functions.export_settings["filedata"]["models"] = {}
	filedata = TF_base_functions.export_settings["filedata"]["models"]

	# get model type
	model_type = model_obj.model_type
	if model_type == "%auto":
		model_type = TF_base_functions.export_settings["model_type"]
	elif model_type == "%inherit" and "parent" in model_object and "model_type" in model_object["parent"]:
		model_type = model_object["parent"]["model_type"]

	file_id = model_type + model_name
	create_model = TF_base_functions.export_settings["subfiles"] or ("forceOverride" in model_object and model_object["forceOverride"])

	if file_id in filedata:
		# model data already exists
		model_data = filedata[file_id]["data"]
	else:
		model_data = {"collider": {"params": { "halfExtents": list((model_obj.tf_boundingInfo.bbMax - model_obj.tf_boundingInfo.bbMin)/2) }, "type": 'BOX'}, "lods": []} # assume a BOX collider (it's very easy to replace this with a MESH collider by hand if that's what's desired, but converting MESH to BOX is more tedious)

		TF_base_functions.export_settings["boundingBox"] = {"bbMax": [0.0, 0.0, 0.0], "bbMin": [0.0, 0.0, 0.0]}

		# get bounding box
		model_data_boundingInfo = {"bbMax": list(model_obj.tf_boundingInfo.bbMax), "bbMin": list(model_obj.tf_boundingInfo.bbMin)}
		if model_data_boundingInfo["bbMax"] != [0,0,0] or model_data_boundingInfo["bbMin"] != [0,0,0]:
			model_data["boundingInfo"] = model_data_boundingInfo

		bounding_box = None
		if TF_base_functions.export_settings["bb"]:
			bounding_box = {"bbMax": [0.0, 0.0, 0.0], "bbMin": [0.0, 0.0, 0.0]}

		has_lods = False
		renderObjects = []
		for lod in model_obj.children:
			if lod.type in {'EMPTY', 'ARMATURE'} and lod.subtype == "lod":
				has_lods = True
				if TF_base_functions.export_settings["game"] == "TPF2":
					lod_data, objects =_get_lod_data_TPF2(lod, lod.children, model_obj, model_type, bounding_box, errors)
				else:
					lod_data, objects =_get_lod_data(lod, lod.children, model_obj, model_type, bounding_box, errors)
				model_data["lods"].append(lod_data)
				renderObjects += objects

		if not has_lods:
			if TF_base_functions.export_settings["game"] == "TPF2":
				lod_data, renderObjects = _get_lod_data_TPF2(None, model_obj.children, model_obj, model_type, bounding_box, errors)
			else:
				lod_data, renderObjects = _get_lod_data(None, model_obj.children, model_obj, model_type, bounding_box, errors)
			model_data["lods"].append(lod_data)

		data = {
			"modelData": model_data,
			"modelObj": model_obj,
		}
		TF_base_functions.call_plugin("save_model_data", data, errors)

		if bounding_box:
			model_data["boundingInfo"] = bounding_box
			
		if TF_base_functions.export_settings["game"] == "TPF2":
			model_data["version"] = 1

		filedata[file_id] = {"data": model_data, "model_type": model_type, "name": model_name}
		if "forceOverride" in model_object:
			filedata[file_id]["forceOverride"] = model_object["forceOverride"]

		UI_BB = TF_base_functions.export_settings["boundingBox"]
		if "boundingInfo" in model_data:
			UI_BB = model_data["boundingInfo"]

		if TF_base_functions.export_settings["ui_icon"]:
			UI_util.create_UI_icons(renderObjects, [{"data": model_data, "model_type": model_type, "name": model_name, "BB": UI_BB, "object": model_obj}], errors)

def write_file(file, errors):
	filename = file["name"] + ".mdl"
	filepath = file["model_type"]

	id = filepath + filename

	resfilepath = "models/model/" + filepath
	resfilename = resfilepath + filename

	if TF_base_functions.export_settings["dir_structure"]:
		osfilepath = TF_base_functions.export_settings["basepath"]+resfilepath
	else:
		osfilepath = TF_base_functions.export_settings["basepath"]

	osfilename = osfilepath + filename

	if (not writer.isfile(osfilename)) or TF_base_functions.export_settings["override"] or ("forceOverride" in file and file["forceOverride"]):
		if not writer.exists(osfilepath):
			writer.makedirs(osfilepath)

		filetext = ""
		# additional code, to do: not restricted to require
		if "__requires" in file["data"]:
			for req in file["data"]["__requires"]:
				filetext += req + '\n'

		filetext += "function data()\nreturn "
		filetext += writer.write_var_struct_string(file["data"], "", 0, True, False)
		filetext += "end"

		data = {
			"text": filetext,
			"modelData": file["data"],
			"fileID": id,
		}
		TF_base_functions.call_plugin("write_model_file", data, errors)

		output = open(osfilename, 'w', encoding="utf-8")
		output.write(data["text"])
		output.close()

	TF_base_functions.export_settings["filelist"].append([osfilename, "res/" + resfilename])

# UI functions
def draw_model(context, layout):
	mydata = context.object

	if mydata:
		boxBB = TF_base_functions.draw_hideable_box(layout, "Bounding Box:", "model_bounding_info", False)
		if boxBB:
			row = boxBB.row()
			row.prop(mydata.tf_boundingInfo, "bbMax", text="BB Max")
			row = boxBB.row()
			row.prop(mydata.tf_boundingInfo, "bbMin", text="BB Min")
			layout.separator()

		TF_base_functions.call_plugin("draw_model_settings_panel", {"layout": layout, "mydata": mydata})
		TF_base_functions.draw_internal_box(mydata, layout, "model")

def draw_lod(context, layout):
	mydata = context.object

	if mydata:
		boxLod = TF_base_functions.draw_hideable_box(layout, "Lod:", "lod", False)
		if boxLod:
			boxLod.prop(mydata.tf_lod, "static", text="Static")
			boxLod.prop(mydata.tf_lod, "visibleFrom", text="Visible from")
			boxLod.prop(mydata.tf_lod, "visibleTo", text="Visible to")
			layout.separator()

		TF_base_functions.call_plugin("draw_lod_settings_panel", {"layout": layout, "mydata": mydata})
		TF_base_functions.draw_internal_box(mydata, layout, "lod")

# Classes
class TFBoundingInfoSetting(bpy.types.PropertyGroup):
	def get_bbMax(self):
		if "bbMax" in self:
			return self["bbMax"]
		elif "bbMax" in self.id_data:
			self["bbMax"] = self.id_data["bbMax"]
			return self.id_data["bbMax"]
		return [0.0, 0.0, 0.0]

	def set_bbMax(self, value):
		self["bbMax"] = value

	def get_bbMin(self):
		if "bbMin" in self:
			return self["bbMin"]
		elif "bbMin" in self.id_data:
			self["bbMin"] = self.id_data["bbMin"]
			return self.id_data["bbMin"]
		return [0.0, 0.0, 0.0]

	def set_bbMin(self, value):
		self["bbMin"] = value

	bbMax = bpy.props.FloatVectorProperty(size=3, default=[0.0, 0.0, 0.0], subtype='XYZ', get=get_bbMax, set=set_bbMax)
	bbMin = bpy.props.FloatVectorProperty(size=3, default=[0.0, 0.0, 0.0], subtype='XYZ', get=get_bbMin, set=set_bbMin)

	@classmethod
	def register(cls):
		bpy.types.Object.tf_boundingInfo = bpy.props.PointerProperty(type=cls)

	@classmethod
	def unregister(cls):
		del bpy.types.Object.tf_boundingInfo

class TFLodSetting(bpy.types.PropertyGroup):
	def get_static(self):
		if "static" in self:
			return self["static"]
		elif "static" in self.id_data:
			self["static"] = self.id_data["static"]
			return self.id_data["static"]
		return False

	def set_static(self, value):
		self["static"] = value

	def get_visibleFrom(self):
		if "visibleFrom" in self:
			return self["visibleFrom"]
		elif "visibleFrom" in self.id_data:
			self["visibleFrom"] = self.id_data["visibleFrom"]
			return self.id_data["visibleFrom"]
		return 0

	def set_visibleFrom(self, value):
		self["visibleFrom"] = value

	def get_visibleTo(self):
		if "visibleTo" in self:
			return self["visibleTo"]
		elif "visibleTo" in self.id_data:
			self["visibleTo"] = self.id_data["visibleTo"]
			return self.id_data["visibleTo"]
		return 2500

	def set_visibleTo(self, value):
		self["visibleTo"] = value

	static = bpy.props.BoolProperty(default=False, get=get_static, set=set_static)
	visibleFrom = bpy.props.IntProperty(default=0, get=get_visibleFrom, set=set_visibleFrom)
	visibleTo = bpy.props.IntProperty(default=2500, get=get_visibleTo, set=set_visibleTo)

	@classmethod
	def register(cls):
		bpy.types.Object.tf_lod = bpy.props.PointerProperty(type=cls)

	@classmethod
	def unregister(cls):
		del bpy.types.Object.tf_lod

__class_list = [
	TFBoundingInfoSetting,
	TFLodSetting,
]

def register():
	for class_type in __class_list:
		bpy.utils.register_class(class_type)
	
	bpy.types.Object.index = bpy.props.IntProperty(default=0)
	
	TF_base_functions.subtype_list.insert(0, ('model', 'Model', '', 0))
	TF_base_functions.subtype_list.append(('lod', 'Lod', '', 2))
	TF_base_functions.subtype_list.append(('node', 'Node', '', 516))
	
	TF_base_functions.export_file_ext["mdl"] = (("mdl", "Model (.mdl)", "", 0))

def unregister():
	for i in range(len(__class_list)-1, -1, -1):
		bpy.utils.unregister_class(__class_list[i])
		
	del bpy.types.Object.index 
	
	TF_base_functions.subtype_list.remove(('model', 'Model', '', 0))
	TF_base_functions.subtype_list.remove(('lod', 'Lod', '', 2))
	TF_base_functions.subtype_list.remove(('node', 'Node', '', 516))
	
	del TF_base_functions.export_file_ext["mdl"]